@@ -19,13 +19,13 @@ use Lemonldap::NG::Portal::Main::Constants qw(
   PE_SAML_IDPSSOINITIATED_NOTALLOWED
   PE_SAML_SESSION_ERROR
   PE_SAML_SIGNATURE_ERROR
-  PE_SAML_SLO_ERROR
+  PE_SLO_ERROR
   PE_SAML_SSO_ERROR
   PE_SAML_UNKNOWN_ENTITY
   PE_SENDRESPONSE
 );
 
-our $VERSION = '2.17.2';
+our $VERSION = '2.21.0';
 
 extends qw(
   Lemonldap::NG::Portal::Main::Auth
@@ -41,6 +41,9 @@ has catch            => ( is => 'rw' );
 use constant sessionKind => 'SAML';
 use constant lsDump      => '_lassoSessionDump';
 use constant liDump      => '_lassoIdentityDump';
+use constant niDump      => '_lassoNameIdDump';
+use constant sIndex      => '_lassoSessionIndex';
+use constant afterData   => 'authFinish';
 
 sub forAuthUser { 'handleAuthRequests' }
 
@@ -99,7 +102,7 @@ sub extractFormInfo {
     my $content_type   = $req->content_type;
 
     # 1.1 SSO assertion consumer
-    if ( $url =~ $self->ssoAssConsumerRe ) {
+    if ( $self->isCallback($req) ) {
         $self->logger->debug(
             "URL $url detected as an SSO assertion consumer URL");
 
@@ -457,6 +460,9 @@ sub extractFormInfo {
         # TODO: call authLogout instead of duplicating SLO
         $req->steps( [ @{ $self->p->beforeLogout }, 'deleteSession' ] );
 
+        # Don't fail on "beforeLogout" errors
+        $req->data->{nofail} = 1;
+
         # Check SAML Message
         my ( $request, $response, $method, $relaystate, $artifact ) =
           $self->checkMessage( $req, $url, $request_method, $content_type,
@@ -475,7 +481,7 @@ sub extractFormInfo {
 
             unless ($result) {
                 $self->userLogger->error("Fail to process logout response");
-                return PE_SAML_SLO_ERROR;
+                return PE_SLO_ERROR;
             }
 
             $self->logger->debug("Logout response is valid");
@@ -531,7 +537,7 @@ sub extractFormInfo {
                 # Logout request was already consumed or is expired
                 $self->userLogger->error(
                     "Message $samlID already used or expired");
-                return PE_SAML_SLO_ERROR;
+                return PE_SLO_ERROR;
             }
 
             # If URL in RelayState, different from portal, redirect user
@@ -542,7 +548,7 @@ sub extractFormInfo {
             }
 
             if (    $req->urldc
-                and $self->conf->{portal} !~ /$req->{urldc}\/?/ )
+                and $req->portal !~ /$req->{urldc}\/?/ )
             {
                 $req->steps( [] );
                 $req->user('TODO');
@@ -565,7 +571,7 @@ sub extractFormInfo {
             # Process logout request
             unless ( $self->processLogoutRequestMsg( $logout, $request ) ) {
                 $self->userLogger->error("Fail to process logout request");
-                return PE_SAML_SLO_ERROR;
+                return PE_SLO_ERROR;
             }
 
             $self->logger->debug("Logout request is valid");
@@ -747,7 +753,7 @@ sub extractFormInfo {
             # Logout response
             unless ( $self->buildLogoutResponseMsg($logout) ) {
                 $self->logger->error("Unable to build SLO response");
-                return PE_SAML_SLO_ERROR;
+                return PE_SLO_ERROR;
             }
 
             # Send response depending on request method
@@ -914,7 +920,7 @@ sub extractFormInfo {
             $req->pdata->{_url} = encode_base64( $req->urldc, '' );
         }
         my $disco_url = $self->conf->{samlDiscoveryProtocolURL};
-        my $portal    = $self->conf->{portal};
+        my $portal    = $req->portal;
         $disco_url .= ( $disco_url =~ /\?/ ? '&' : '?' )
           . build_urlencoded(
             entityID      => $self->getMetaDataURL( 'samlEntityID', 0, 1 ),
@@ -954,15 +960,14 @@ sub extractFormInfo {
         return $urlcheck unless ( $urlcheck == PE_OK );
 
         # IDP list
-        my @list       = ();
-        my $portalPath = $self->conf->{portal};
+        my @list = ();
 
         foreach ( keys %{ $self->idpList } ) {
             my $idpName = $self->{idpList}->{$_}->{name};
             $idpName = $self->{idpList}->{$_}->{displayName}
               if $self->{idpList}->{$_}->{displayName};
             my $icon    = $self->{idpList}->{$_}->{icon};
-            my $order   = $self->{idpList}->{$_}->{order} // 0;
+            my $order   = $self->{idpList}->{$_}->{order} // 999999;
             my $tooltip = $self->{idpList}->{$_}->{tooltip} || $idpName;
             my $img_src = '';
 
@@ -970,7 +975,7 @@ sub extractFormInfo {
                 $img_src =
                   ( $icon =~ m#^https?://# )
                   ? $icon
-                  : $portalPath . $self->p->staticPrefix . "/common/" . $icon;
+                  : $self->p->staticPrefix . "/common/" . $icon;
             }
             $self->logger->debug( "IDP "
                   . $self->{idpList}->{$_}->{name}
@@ -1156,9 +1161,11 @@ sub authenticate {
 
 sub setAuthSessionInfo {
     my ( $self, $req ) = @_;
-    my $login      = $req->data->{_lassoLogin};
-    my $idp        = $req->data->{_idp};
-    my $idpConfKey = $req->data->{_idpConfKey};
+    my $login          = $req->data->{_lassoLogin};
+    my $idp            = $req->data->{_idp};
+    my $idpConfKey     = $req->data->{_idpConfKey};
+    my $nameIdentifier = $req->data->{_nameID};
+    my $session_index  = $req->data->{_sessionIndex};
 
     # Get SAML assertion
     my $assertion = $self->getAssertion($login);
@@ -1244,6 +1251,9 @@ sub setAuthSessionInfo {
     # Dump Lasso objects in session
     $req->{sessionInfo}->{ $self->lsDump } = $session->dump()  if $session;
     $req->{sessionInfo}->{ $self->liDump } = $identity->dump() if $identity;
+    $req->{sessionInfo}->{ $self->niDump } = $nameIdentifier->dump()
+      if ref($nameIdentifier);
+    $req->{sessionInfo}->{ $self->sIndex } = $session_index if $session_index;
 
     # Keep SAML Token in session
     my $store_samlToken =
@@ -1257,26 +1267,27 @@ sub setAuthSessionInfo {
     }
 
     $req->data->{_lassoLogin} = $login;
-    push @{ $req->steps }, sub { $self->authFinish(@_) };
 
     return PE_OK;
 }
 
-# Inserted in $req->steps by authenticate()
+# Saves the link between IDP-side Session ID/NameID and LLNG session
 sub authFinish {
     my ( $self, $req ) = @_;
-    my %h;
+
+    # Get saved Lasso objects
+    my $nameid        = $req->sessionInfo->{ $self->niDump };
+    my $session_index = $req->sessionInfo->{ $self->sIndex };
+
+    # Auth::SAML was not used for this session
+    return unless ($nameid);
 
     # Real session was stored, get id and utime
     my $id    = $req->{id};
     my $utime = $req->{sessionInfo}->{_utime};
 
-    # Get saved Lasso objects
-    my $nameid        = $req->data->{_nameID};
-    my $session_index = $req->data->{_sessionIndex};
-
     $self->logger->debug( "Store NameID "
-          . $nameid->dump
+          . $nameid
           . ( $session_index ? " and SessionIndex $session_index" : "" )
           . " for session $id" );
 
@@ -1284,7 +1295,7 @@ sub authFinish {
     $infos->{type}          = 'saml';            # Session type
     $infos->{_utime}        = $utime;            # Creation time
     $infos->{_saml_id}      = $id;               # SSO session id
-    $infos->{_nameID}       = $nameid->dump;     # SAML NameID
+    $infos->{_nameID}       = $nameid;           # SAML NameID
     $infos->{_sessionIndex} = $session_index;    # SAML SessionIndex
 
     # Save SAML session
@@ -1315,7 +1326,7 @@ sub authLogout {
 
     unless ($session_dump) {
         $self->userLogger->error("Could not get session dump from session");
-        return PE_SAML_SLO_ERROR;
+        return PE_SLO_ERROR;
     }
 
     # IDP HTTP method
@@ -1350,7 +1361,7 @@ sub authLogout {
         $method, $signSLOMessage );
     unless ($logout) {
         $self->logger->error("Could not create logout request");
-        return PE_SAML_SLO_ERROR;
+        return PE_SLO_ERROR;
     }
 
     $self->logger->debug("Logout request created");
@@ -1358,7 +1369,7 @@ sub authLogout {
     # Keep request ID in memory to prevent replay
     unless ( $self->storeReplayProtection( $logout->request()->ID ) ) {
         $self->logger->error("Unable to store Logout request ID");
-        return PE_SAML_SLO_ERROR;
+        return PE_SLO_ERROR;
     }
 
     # Send request depending on request method
@@ -1413,7 +1424,7 @@ sub authLogout {
 
         unless ($response) {
             $self->logger->error("No logout response to SOAP request");
-            return PE_SAML_SLO_ERROR;
+            return PE_SLO_ERROR;
         }
 
         # Create Logout object
@@ -1433,7 +1444,7 @@ sub authLogout {
 
         unless ($result) {
             $self->logger->error("Fail to process logout response");
-            return PE_SAML_SLO_ERROR;
+            return PE_SLO_ERROR;
         }
 
         $self->logger->debug("Logout response is valid");
@@ -1445,17 +1456,22 @@ sub authLogout {
 
             # Logout request was already consumed or is expired
             $self->userLogger->error("Message $samlID already used or expired");
-            return PE_SAML_SLO_ERROR;
+            return PE_SLO_ERROR;
         }
 
         return PE_OK;
     }
     else {
         $self->userLogger->error("Lasso method $method not implemented here");
-        return PE_SAML_SLO_ERROR;
+        return PE_SLO_ERROR;
     }
 }
 
+sub isCallback {
+    my ( $self, $req ) = @_;
+    return $req->uri =~ $self->ssoAssConsumerRe;
+}
+
 sub handleAuthRequests {
     my ( $self, $req ) = @_;
     if ( $req->uri =~ $self->sloRe ) {
