@@ -4,10 +4,13 @@ use strict;
 use Mouse;
 use URI::Escape;
 use Lemonldap::NG::Common::FormEncode;
-use Lemonldap::NG::Portal::Main::Constants
-  qw(PE_OK PE_BADURL PE_GET_SERVICE_NOT_ALLOWED URIRE);
+use Lemonldap::NG::Portal::Main::Constants qw(
+  PE_OK
+  PE_UNAUTHORIZEDURL
+  PE_GET_SERVICE_NOT_ALLOWED URIRE
+);
 
-our $VERSION = '2.0.12';
+our $VERSION = '2.0.15';
 
 extends 'Lemonldap::NG::Portal::Main::Issuer';
 
@@ -29,7 +32,7 @@ sub init {
         return 0;
     }
     $self->{rule} = $rule;
-    return 0 unless ( $self->SUPER::init() );
+    return 0 unless $self->SUPER::init();
     return 1;
 }
 
@@ -49,14 +52,8 @@ sub run {
 
     # Session creation timestamp
     my $time = $req->{sessionInfo}->{_utime} || time();
-    $req->path =~ m#^$self->{conf}->{issuerDBGetPath}/(log(?:in|out))#;
-    my $logInOut = $1 || 'login';
-    if ( $logInOut eq 'login' ) {
-        $self->logger->debug("IssuerGet: request for login");
-        $self->computeGetParams($req);
-        return PE_OK;
-    }
-    elsif ( $logInOut eq 'logout' ) {
+
+    if ( $req->path =~ m#^$self->{conf}->{issuerDBGetPath}/logout# ) {
         $self->logger->debug("IssuerGet: request for logout");
 
         # TODO
@@ -64,14 +61,15 @@ sub run {
         return PE_OK;
     }
     else {
-        $self->logger->error("IssuerGet: bad url");
-        return PE_BADURL;
+        $self->logger->debug("IssuerGet: request for login");
+        $self->computeGetParams($req);
+        return PE_OK;
     }
 }
 
 # Nothing to do here for now
 sub logout {
-    PE_OK;
+    return PE_OK;
 }
 
 # INTERNAL METHODS
